/*
jViewBox 2.0 beta

COPYRIGHT NOTICE
Copyright (c) 2002  Jack C. Wei, Scott C. Neu, and Daniel J. Valentino

See README.license for license notices.
 */

import java.awt.BorderLayout;
import java.awt.Image;
import java.awt.MediaTracker;
import java.awt.Toolkit;
import java.io.File;
import java.util.Arrays;
import java.util.List;
import java.util.Vector;
import javax.swing.JFileChooser;
import javax.swing.JFrame;

import org.medtoolbox.jviewbox.viewport.ViewportGrid;
import org.medtoolbox.jviewbox.viewport.ViewportToolBar;

import tools.GlobalizeTool;
import tools.LayoutTool;
import tools.PanTool;
import tools.Rotate90Tool;
import tools.ScrollTool;
import tools.UndoTool;
import tools.WindowLevelTool;
import tools.ZoomTool;

/**
 * Sample application to view a set of medical images.  The viewer reads
 * images from the file names provided on the command line as arguments
 * and displays them.
 *
 * @version January 8, 2004
 */
public class Viewer extends JFrame
{
    /**
     * Constructs a Viewer to display the specified images.
     *
     * @param images Images to display.
     */
    public Viewer(List images)
    {
	super("Viewer");

	// Make sure all images are fully loaded
	MediaTracker tracker = new MediaTracker(this);
	for (int i = 0; i < images.size(); i++) {
	    tracker.addImage((Image)images.get(i), i);
	}
	System.out.print("Loading images... ");
	System.out.flush();
	while (!tracker.checkAll()) {
	    try {
		tracker.waitForAll();
	    }
	    catch (InterruptedException e) {}
	}
	System.out.println("Done.");

	// Remove any images failed to load
	if (tracker.isErrorAny()) {
	    System.out.println("Unable to load one or more image files.");
	    images = new Vector(images);
	    images.removeAll(Arrays.asList(tracker.getErrorsAny()));
	}

	// Create a ToolBar and add Tools to it
	ViewportToolBar toolBar = new ViewportToolBar();
	toolBar.addTool(new PanTool(), "Pan");
	toolBar.addTool(new ZoomTool(), "Zoom");
	toolBar.addTool(new WindowLevelTool(), "Window/Level");
	toolBar.addTool(new Rotate90Tool(), "Rotate");
	toolBar.addTool(new UndoTool(), "Undo");
	toolBar.addTool(new LayoutTool(), "Layout");
	toolBar.addTool(new ScrollTool(), "Scroll");
	toolBar.addTool(new GlobalizeTool(), "Globalize");

	// Create a ViewportGrid to display the images
	ViewportGrid grid = new ViewportGrid(images);

	// Link the ToolBar to the ViewportGrid
	grid.registerViewportToolBar(toolBar);

	// Arrange the ToolBar and ViewportGrid in this frame
	getContentPane().add(toolBar, BorderLayout.NORTH);
	getContentPane().add(grid, BorderLayout.CENTER);
    }

    /**
     * Application main method.
     *
     * @param args Command-line arguments as image file names to view.
     */
    public static void main(String[] args) 
    {
	// If there is no command-line arguments, pop up a JFileChooser and
	// let the user choose image files to load
	if (args.length == 0) {
	    JFileChooser chooser = new JFileChooser(".");
	    chooser.setFileHidingEnabled(true);
	    chooser.setFileSelectionMode(JFileChooser.FILES_ONLY);
	    chooser.setMultiSelectionEnabled(true);
	    if (chooser.showOpenDialog(null) == JFileChooser.APPROVE_OPTION) {
		File[] files = chooser.getSelectedFiles();
		args = new String[files.length];
		for (int i = 0; i < files.length; i++) {
		    args[i] = files[i].getPath();
		}
	    }
	}

	// Read the image files using Toolkit
	Toolkit toolkit = Toolkit.getDefaultToolkit();
	Vector allImages = new Vector();
	for (int i = 0; i < args.length; i++) {
	    Image image = toolkit.getImage(args[i]);
	    if (image != null) {
		allImages.add(image);
	    }
	}

	// Check for at least one image to show
	if (allImages.isEmpty()) {
	    System.out.println("No image to show.");
	    System.exit(1);
	}

	// Create a Viewer and make it visible
	Viewer viewer = new Viewer(allImages);
	viewer.setDefaultCloseOperation(EXIT_ON_CLOSE);
	viewer.setSize(800, 600);
	viewer.setVisible(true);
    }
}
