/*
jViewBox 2.0 beta

COPYRIGHT NOTICE
Copyright (c) 2002  Jack C. Wei, Scott C. Neu, and Daniel J. Valentino

See README.license for license notices.
 */

import java.awt.BorderLayout;
import java.io.File;
import java.io.IOException;
import java.util.Iterator;
import java.util.List;
import java.util.Vector;
import javax.swing.JFileChooser;
import javax.swing.JFrame;

import javax.imageio.ImageIO;
import javax.imageio.ImageReader;
import javax.imageio.stream.FileImageInputStream;

import org.medtoolbox.jviewbox.imagesource.ImageReaderAdapter;
import org.medtoolbox.jviewbox.viewport.ViewportGrid;
import org.medtoolbox.jviewbox.viewport.ViewportToolBar;

import tools.GlobalizeTool;
import tools.LayoutTool;
import tools.PanTool;
import tools.Rotate90Tool;
import tools.ScrollTool;
import tools.UndoTool;
import tools.WindowLevelTool;
import tools.ZoomTool;

/**
 * Sample application to view a set of medical images.  The viewer reads
 * images from the file names provided on the command line as arguments
 * and displays them.
 *
 * @version January 8, 2004
 */
public class Viewer2 extends JFrame
{
    /**
     * Constructs a Viewer to display the specified images.
     *
     * @param images Images to display.
     */
    public Viewer2(List images)
    {
	super("Viewer");

	// Create a ToolBar and add Tools to it
	ViewportToolBar toolBar = new ViewportToolBar();
	toolBar.addTool(new PanTool(), "Pan");
	toolBar.addTool(new ZoomTool(), "Zoom");
	toolBar.addTool(new WindowLevelTool(), "Window/Level");
	toolBar.addTool(new Rotate90Tool(), "Rotate");
	toolBar.addTool(new UndoTool(), "Undo");
	toolBar.addTool(new LayoutTool(), "Layout");
	toolBar.addTool(new ScrollTool(), "Scroll");
	toolBar.addTool(new GlobalizeTool(), "Globalize");

	// Create a ViewportGrid to display the images
	ViewportGrid grid = new ViewportGrid(images);

	// Link the ToolBar to the ViewportGrid
	grid.registerViewportToolBar(toolBar);

	// Arrange the ToolBar and ViewportGrid in this frame
	getContentPane().add(toolBar, BorderLayout.NORTH);
	getContentPane().add(grid, BorderLayout.CENTER);
    }

    /**
     * Application main method.
     *
     * @param args Command-line arguments as image file names to view.
     */
    public static void main(String[] args) 
    {
	// If there is no command-line arguments, pop up a JFileChooser and
	// let the user choose image files to load
	File[] files;
	if (args.length == 0) {
	    JFileChooser chooser = new JFileChooser(".");
	    chooser.setFileHidingEnabled(true);
	    chooser.setFileSelectionMode(JFileChooser.FILES_ONLY);
	    chooser.setMultiSelectionEnabled(true);
	    if (chooser.showOpenDialog(null) == JFileChooser.APPROVE_OPTION) {
		files = chooser.getSelectedFiles();
	    }
	    // The user cancelled the chooser
	    else {
		files = new File[0];
	    }
	}

	// Use command-line arguments
	else {
	    files = new File[args.length];
	    for (int i = 0; i < args.length; i++) {
		files[i] = new File(args[i]);
	    }
	}

	// Read the image files using Image I/O plugins
	Vector allImages = new Vector();
	for (int i = 0; i < files.length; i++) {
	    try {
		// 1st try: use FileImageInputStream as input
		Object input = new FileImageInputStream(files[i]);
		Iterator it = ImageIO.getImageReaders(input);

		// Found no reader at the 1st try?
		if (!it.hasNext()) {
		    // 2nd try: use the file itself as input
		    input = files[i];
		    it = ImageIO.getImageReaders(files[i]);

		    // Found no reader at the 2nd try?
		    if (!it.hasNext()) {
			// 3rd try: get a reader from the file name suffix
			String fileName = files[i].getName();
			int lastDot = fileName.lastIndexOf('.');
			if (lastDot >= 0) {
			    String suffix = fileName.substring(lastDot + 1);
			    it = ImageIO.getImageReadersBySuffix(suffix);
			}
		    }
		}

		// Found any reader at all?
		if (it.hasNext()) {
		    ImageReader reader = (ImageReader)it.next();
		    reader.setInput(input, false);

		    // Wrap each image in the reader in an adapter for jViewBox
		    int numImages = reader.getNumImages(true);
		    for (int j = 0; j < numImages; j++) {
			allImages.add(new ImageReaderAdapter(reader, j));
		    }
		}

		// No reader available for the file
		else {
		    System.out.println("Unable to find an image decoder for " +
				       files[i]);
		    System.out.println("Skipping the file");
		}
	    }
	    catch (IOException e) {
		System.out.println("Unable to read " + files[i] + " due to " +
				   e);
		System.out.println("Skipping the file");
	    }
	}

	// Check for at least one image to show
	if (allImages.isEmpty()) {
	    System.out.println("No image to show");
	    System.exit(1);
	}

	// Create a Viewer and make it visible
	Viewer2 viewer = new Viewer2(allImages);
	viewer.setDefaultCloseOperation(EXIT_ON_CLOSE);
	viewer.setSize(800, 600);
	viewer.setVisible(true);
    }
}
