/*
jViewBox 2.0 beta

COPYRIGHT NOTICE
Copyright (c) 2003  Jack C. Wei, Scott C. Neu, and Daniel J. Valentino

See README.license for license notices.
 */

package tools;

import java.awt.event.MouseEvent;
import org.medtoolbox.jviewbox.viewport.Viewport;
import org.medtoolbox.jviewbox.viewport.ViewportGrid;
import org.medtoolbox.jviewbox.viewport.ViewportCluster;
import org.medtoolbox.jviewbox.viewport.ViewportTool;

/**
 * Tool for cycling through different Viewport layouts of a ViewportGrid.
 * This tool only works on ViewportGrids.
 * <p>
 * When the mouse cursor is clicked on a ViewportGrid, the Viewports inside
 * of it will be rearranged into a different layout in a predefined sequence.
 * Left button click advances to the next layout while the other button(s) 
 * backtracks to the previous layout. When either end of the sequence is
 * reached, this tool cycles to the other end.
 *
 * @version January 8, 2004
 */
public class LayoutTool extends ViewportTool
{
    /**
     * Predefined layout sequence. It is in the form of
     * int[sequence length][2] where the [i][0] element is the width (# of
     * columns) and [i][1] the height (# of rows) of the i-th layout.
     */
    private final int[][] _layouts;

    /**
     * Constructs a LayoutTool with a default layout sequence of { 1x1, 2x2,
     * 4x4 }.
     */
    public LayoutTool()
    {
	this(new int[] { 1, 2, 4 });
    }

    /**
     * Constructs a LayoutTool with the specified layout sequence.
     *
     * @param layouts Sequence of integers each of which gives the equal width
     *		      (# of columns) and height (# of rows) of one layout in
     *		      the sequence.
     *
     * @throws NullPointerException if <code>layouts</code> is
     *	       <code>null</code>.
     * @throws IllegalArgumentException if <code>layouts</code> is empty
     *	       (zero-lengthed).
     */
    public LayoutTool(int[] layouts)
    {
	this(_createSquareLayout(layouts));
    }

    /**
     * Constructs an int[][] square layout sequence out of the specified
     * dimension sequence.
     */
    private static int[][] _createSquareLayout(int[] layouts)
    {
	int[][] l = new int[layouts.length][2];
	for (int i = 0; i < layouts.length; i++) {
	    l[i][0] = l[i][1] = layouts[i];
	}
	return l;
    }

    /**
     * Constructs a LayoutTool with the specified layout sequence.
     *
     * @param layouts Array in the form of int[sequence length][2] where the
     *		      [i][0] element is the width (# of columns) and [i][1]
     *		      the height (# of rows) of the i-th layout.
     *
     * @throws NullPointerException if <code>layouts</code> is
     *	       <code>null</code>.
     * @throws IllegalArgumentException if <code>layouts</code> is empty
     *	       (zero-lengthed).
     */
    public LayoutTool(int[][] layouts)
    {
	super("Layout Tool", "Next Layout", "Previous Layout",
	      "Previous Layout");

	if (layouts.length == 0) {
	    throw new IllegalArgumentException("layouts cannot be empty.");
	}
	_layouts = new int[layouts.length][2];
	for (int i = 0; i < layouts.length; i++) {
	    _layouts[i][0] = layouts[i][0];
	    _layouts[i][1] = layouts[i][1];
	}
    }

    /**
     * Invoked when a mouse button is clicked (pressed and released without
     * movement in between) on a ViewportCluster.
     *
     * @param vpc ViewportCluster which received the event.
     * @param vp Viewport which the mouse cursor was on during the event;
     *		 <code>null</code> if the mouse was outside all visible
     *		 Viewports in the ViewportCluster.
     * @param e Event created by the button click.
     * @param button The button that was clicked.
     */
    public void mouseClicked(ViewportCluster vpc, Viewport vp, MouseEvent e, 
			     int button)
    {
	if (vpc != null && vpc instanceof ViewportGrid) {
	    ViewportGrid vpg = (ViewportGrid)vpc;
	    int rows = vpg.getNumberOfRows();
	    int cols = vpg.getNumberOfColumns();

	    // Look for the current layout in the sequence
	    // index will be _layouts.length if not found
	    int index = 0;
	    for (; index < _layouts.length; index++) {
		if (cols == _layouts[index][0] && rows == _layouts[index][1]) {
		    break;
		}
	    }

	    // Left button advances to the next layout
	    // Note that if the current layout is not found, the first layout
	    // in the sequence will be chosen
	    if (button == LEFT_MOUSE_BUTTON) {
		index++;
		if (index >= _layouts.length) {
		    index = 0;
		}
	    }

	    // Other buttons backtracks to the previous layout
	    // Note that if the current layout is not found, the last layout
	    // in the sequence will be chosen
	    else {
		index--;
		if (index < 0) {
		    index = _layouts.length - 1;
		}
	    }

	    // Apply the new layout
	    vpg.setGrid(_layouts[index][1], _layouts[index][0]);

	    // Repaint the ViewportGrid
	    vpg.repaint();
	}
    }
}
