/*
jViewBox 2.0 beta

COPYRIGHT NOTICE
Copyright (c) 2003  Jack C. Wei, Scott C. Neu, and Daniel J. Valentino

See README.license for license notices.
 */

package tools;

import java.awt.Point;
import java.awt.image.BufferedImage;
import java.awt.event.MouseEvent;
import org.medtoolbox.jviewbox.BufferedImageUtilities;
import org.medtoolbox.jviewbox.LookUpTable;
import org.medtoolbox.jviewbox.LinearLookUpTable;
import org.medtoolbox.jviewbox.viewport.Viewport;
import org.medtoolbox.jviewbox.viewport.ViewportCluster;
import org.medtoolbox.jviewbox.viewport.ViewportTool;

/**
 * Tool for changing the window and level values of the LinearLookUpTable
 * applied to the image inside a Viewport. This tool only works on Viewports
 * with LinearLookUpTables.
 * <p>
 * When the mouse cursor is pressed and dragged in a Viewport, its horizontal
 * movement affects the level value and the vertical movement affects the
 * window value. A drag to the left decreases the level value while a drag to
 * the right increases it. An upward drag decreases the window value while a
 * downward drag increases it.
 * <p>
 * When the mouse is dragged with the left button down, the normal sensitivity
 * applies. If the mouse is dragged with any other button down, a 5x finer
 * sensitivity applies (i.e., the same amount of mouse movement translates to
 * one fifth of change in window and level values.)
 * <p>
 * If the mouse is double-clicked on a Viewport whose image is a 
 * <code>BufferedImage</code>, this tool calls
 * <code>BufferedImageUtilities.createLUTByScanning</code> to create a look-up
 * table to apply to the image.
 *
 * @version January 8, 2004
 */
public class WindowLevelTool extends ViewportTool
{
    /**
     * Mouse movement in pixels that translates to the full scale of the 
     * LinearLookUpTable being adjusted.
     */
    private int _sensitivity;

    /** Original window value. */
    private int _origWindow;

    /** Original level value. */
    private int _origLevel;

    /** Anchor point relative to which window and level are calculated. */
    private final Point _anchor = new Point();

    /** Viewport whose LinearLookUpTable are being adjusted. */
    private Viewport _vp = null;

    /** LinearLookUpTable whose window/level values are being adjusted. */
    private LinearLookUpTable _vll = null;

    /**
     * Constructs a WindowLevelTool with a default sensitivity of 1000 pixels
     * to the full scale of a LinearLookUpTable.
     */
    public WindowLevelTool()
    {
	this(1000);
    }

    /**
     * Constructs a WindowLevelTool with the specified mouse window/level
     * sensitivity.
     *
     * @param sensitivity Mouse movement in pixels that translates
     *			  to the full scale of the LinearLookUpTable being
     *			  adjusted.
     *
     * @throws IllegalArgumentException If the window/level sensitivity
     *         is invalid.
     */
    public WindowLevelTool(int sensitivity)
    {
	super("Window/Level Tool", "Window/Level an Image", 
	      "Finer Window/Level", "Finer Window/Level");

	if (sensitivity <= 0) {
	    throw new IllegalArgumentException("WindowLevelTool: A window/" +
					       "level sensitivity of " +
					       sensitivity +
					       " is not allowed in the " +
					       "constructor.");
	}

	_sensitivity = sensitivity;
    }


    /**
     * Invoked when a mouse button is pressed on a ViewportCluster.
     *
     * @param vpc ViewportCluster which received the event.
     * @param vp Viewport which the mouse cursor was on during the event;
     *		 <code>null</code> if the mouse was outside all visible
     *		 Viewports in the ViewportCluster.
     * @param e Event created by the button press.
     * @param button The button that was pressed.
     */
    public void mousePressed(ViewportCluster vpc, Viewport vp, MouseEvent e, 
			     int button)
    {
	// Determine if there is a Linear LUT to window/level
	if (vp != null && vp.getLut() instanceof LinearLookUpTable) {
	    _vp = vp;

	    // Save the current state
	    vp.saveCurrentState();

	    // Set the anchor and original window/level values
	    _anchor.x = e.getX();
	    _anchor.y = e.getY();
	    _vll = (LinearLookUpTable)vp.getLut();
	    _origWindow = _vll.getWindow();
	    _origLevel = _vll.getLevel();

	}
	else {
	    _vp = null;
	    _vll = null;
	}
    }

    /**
     * Invoked when a mouse button is pressed and then the mouse is dragged
     * on a ViewportCluster.
     *
     * @param vpc ViewportCluster which received the event.
     * @param vp Viewport which the mouse cursor was on during the event.
     * @param e Event created by the button drag.
     * @param button The button that was pressed and dragged.
     */
    public void mouseDragged(ViewportCluster vpc, Viewport vp, MouseEvent e, 
			     int button)
    {
	if (_vll != null && _vp != null && vpc != null) {

	    // Find out the real sensitivity
	    double sensitivity = (double)_sensitivity / _vll.getSize();
	    if (button != LEFT_MOUSE_BUTTON) {
		sensitivity *= 5.0;
	    }

	    // Set the new window and level values
	    int window = (int)(_origWindow + 
			       (e.getY() - _anchor.y) / sensitivity);
	    int level = (int)(_origLevel + 
			      (e.getX() - _anchor.x) / sensitivity);
	    _vll.setWindowLevel(window, level);

	    vpc.repaint(_vp);

	}
    }

    /**
     * Invoked when a mouse button is clicked (pressed and released without
     * movement in between) on a ViewportCluster. Override this method to
     * implement tool functions triggered by the specific event.
     *
     * @param vpc ViewportCluster which received the event.
     * @param vp Viewport which the mouse cursor was on during the event;
     *		 <code>null</code> if the mouse was outside all visible
     *		 Viewports in the ViewportCluster.
     * @param e Event created by the button click.
     * @param button The button that was clicked.
     */
    public void mouseClicked(ViewportCluster vpc, Viewport vp, MouseEvent e, 
			     int button)
    {
	// Respond to double click only
	if (vp != null && e.getClickCount() == 2) {

	    // Can only handle BufferedImage
	    if (vp.getLut() instanceof LinearLookUpTable &&
		vp.getImage() instanceof BufferedImage) {

		// Save the current state
		vp.saveCurrentState();

		// Create a LUT by scanning the image for min/max
		BufferedImage image = (BufferedImage)vp.getImage();
		LookUpTable lut =
		    BufferedImageUtilities.createLUTByScanning(image);
		if (lut != null) {
		    vp.setLut(lut);
		    vpc.repaint(vp);
		}
	    }
	}
    }

}
