/*
jViewBox 2.0 beta

COPYRIGHT NOTICE
Copyright (c) 2003  Jack C. Wei, Scott C. Neu, and Daniel J. Valentino

See README.license for license notices.
*/

package tools;

import java.awt.Point;
import java.awt.event.MouseEvent;
import org.medtoolbox.jviewbox.viewport.Viewport;
import org.medtoolbox.jviewbox.viewport.ViewportCluster;
import org.medtoolbox.jviewbox.viewport.ViewportTool;

/**
 * Tool for scaling an image inside a Viewport. This tool calculates and
 * updates the image's pan offset so that the zooming is centered at the
 * anchor point, i.e., where the mouse button is first pressed.
 * <p>
 * When the mouse cursor is pressed and dragged in a Viewport, its vertical
 * movement affects the scale factor applied to the image in the Viewport.
 * An upward drag decreases the scale factor (minifying, zooming out) while a
 * downward drag increases the scale factor (magnifying, zooming in).
 * <p>
 * When the mouse is dragged with the left button down, the normal sensitivity
 * applies. If the mouse is dragged with any other button down, a 5x finer
 * sensitivity applies (i.e., the same amount of vertical mouse movement
 * translates to one fifth of change in scale factors.)
 * <p>
 * Maximum and minimum limits apply to the scale factor. The limits can be
 * specified in the constructor.
 *
 * @version January 8, 2004
 */
public class ZoomTool extends ViewportTool
{
    /**
     * Mouse movement in pixels that translates to a change of 1.0 to the
     * scale factors.
     */
    private int _sensitivity;

    /** Minimum allowed value for the Y scale factor. */
    private double _minimumScale;

    /** Maximum allowed value for the Y scale factor. */
    private double _maximumScale;

    /** Original X scale factor. */
    private double _origScaleX;

    /** Original Y scale factor. */
    private double _origScaleY;

    /**
     * Location of the mouse cursor when a mouse button is first pressed
     * inside a ViewportCluster.
     */
    private final Point _anchor = new Point();

    /** Used for calculating the compensational panning. */
    private final Point _r = new Point();

    /** Used for calculating the compensational panning. */
    private final Point _p = new Point();

    /** Viewport in which zooming is underway. */
    private Viewport _vp;

    /** ViewportCluster in which zooming is underway. */
    private ViewportCluster _vpc;

    /** Constructs a ZoomTool with default settings. */
    public ZoomTool()
    {
	this(200);
    }

    /**
     * Constructs a ZoomTool with the specified sensitivity and default
     * upper and lower bounds on the scale factor.
     */
    public ZoomTool(int sensitivity)
    {
	this(sensitivity, 0.01, 5.0);
    }

    /**
     * Constructs a ZoomTool with the specified sensitivity and upper/lower
     * bounds on the scale factor.
     *
     * @param sensitivity Mouse movement in pixels that translates to a change
     *			  of 1.0 to the scale factors.
     * @param minimumScale Minimum value of the Y scale factor allowed.
     * @param maximumScale Maximum value of the Y scale factor allowed.
     *
     * @throws IllegalArgumentException If any of the scale sensitivity,
     *         minimum scale, and maximum scale is invalid.
     */
    public ZoomTool(int sensitivity, double minimumScale, double maximumScale)
    {
	super("Zoom Tool", "Zoom an Image", "", "");

	// Check bounds
	if (sensitivity <= 0 || minimumScale < 0.0001 ||
	    maximumScale < 0.0001) {
	    throw new IllegalArgumentException("ZoomTool: A scale sensitivity"+
					       " of " + sensitivity +
					       " and a minimum scale of " +
					       minimumScale + " and a maximum"+
					       " scale of " + maximumScale +
					       " are not allowed in the " +
					       "constructor.");
	}

	_sensitivity = sensitivity;
	_minimumScale = minimumScale;
	_maximumScale = maximumScale;
    }

    /**
     * Invoked when a mouse button is pressed on a ViewportCluster.
     *
     * @param vpc ViewportCluster which received the event.
     * @param vp Viewport which the mouse cursor was on during the event.
     * @param e Event created by the button press.
     * @param button The button that was pressed.
     */
    public void mousePressed(ViewportCluster vpc, Viewport vp, MouseEvent e, 
			     int button)
    {
	_vp = vp;
	_vpc = vpc;

	// Set the anchor and current scale
	if (_vp != null) {
	    _anchor.x = e.getX();
	    _anchor.y = e.getY();
	    _p.x = _vp.getPanX();
	    _p.y = _vp.getPanY();
	    _r.x = _anchor.x - _vp.getX() - _p.x;
	    _r.y = _anchor.y - _vp.getY() - _p.y;
	    _origScaleX = _vp.getScaleX();
	    _origScaleY = _vp.getScaleY();

	    // Save the current state
	    _vp.saveCurrentState();
	}
    }

    /**
     * Invoked when a mouse button is pressed and then the mouse is dragged
     * on a ViewportCluster.
     *
     * @param vpc ViewportCluster which received the event.
     * @param vp Viewport which the mouse cursor was on during the event.
     * @param e Event created by the button drag.
     * @param button The button that was pressed and dragged.
     */
    public void mouseDragged(ViewportCluster vpc, Viewport vp, MouseEvent e, 
			     int button)
    {
	if (_vp != null && _vpc != null) {

	    // Compute the change in the Y scale factors
	    double deltaScaleY = (double)(e.getY() - _anchor.y) /
		(button == LEFT_MOUSE_BUTTON ? _sensitivity : _sensitivity*5);

	    // Compute the new scale factors
	    double scaleY = _origScaleY + deltaScaleY;
	    if (scaleY < _minimumScale) {
		scaleY = _minimumScale;
	    }
	    else if (scaleY > _maximumScale) {
		scaleY = _maximumScale;
	    }
	    double scaleX = scaleY * _origScaleX / _origScaleY;

	    // Calculate new pan
	    double rX = _r.x * scaleX / _origScaleX;
	    double rY = _r.y * scaleY / _origScaleY;
	    _vp.setPan((int)(_p.x + _r.x - rX), (int)(_p.y + _r.y - rY));
	
	    _vp.setScale(scaleX, scaleY);
	    _vpc.repaint(_vp);
	}
    }

    /**
     * Invoked when a mouse button is released on a ViewportCluster. Override
     * this method to implement tool functions triggered by the specific event.
     *
     * @param vpc ViewportCluster which received the event.
     * @param vp Viewport which the mouse cursor was on during the event;
     *		 <code>null</code> if the mouse was outside all visible
     *		 Viewports in the ViewportCluster.
     * @param e Event created by the button release.
     * @param button The button that was released.
     */
    public void mouseReleased(ViewportCluster vpc, Viewport vp, MouseEvent e, 
			      int button)
    {
	_vp = null;
	_vpc = null;
    }
}
